#!/usr/bin/env python

"""DISTRIBUTED SGD"""

import datetime
import os
from time import time

import torch.distributed as dist
import torch.optim as optim
from torch.autograd import Variable

# local imports
from config import args
from utils import average_gradients
from utils import calculate_loss_value
from utils import calculate_model_loss
from utils import init_dataset
from utils import init_model
from utils import test_dataset


def dsgd(rank, size):
  """ Distributed Synchronous SGD Example """
  type = args.model_type
  desc = 'dsgd'

  output_dir = '../output_{}_{}'.format(args.dataset, type)
  if not os.path.exists(output_dir):
    os.mkdir(output_dir)

  # -----------------------------------
  #              DATASET
  # -----------------------------------
  train_set, bsz = init_dataset(args.dataset,
                                args.batch_size,
                                is_train=True,
                                rank=rank)
  test_set, bsz = init_dataset(args.dataset,
                               args.batch_size,
                               is_train=False,
                               rank=rank)

  # -----------------------------------
  #              MODEL
  # -----------------------------------
  model = init_model(args.dataset)

  # -----------------------------------
  #        CALCULATE INIT LOSS
  # -----------------------------------
  calculate_loss_value(rank, size, model, train_set,
                       batch_size=bsz, master_rank=0,
                       type=type, desc=desc, output_dir=output_dir)
  # -----------------------------------

  optimizer = optim.SGD(model.parameters(),
                        lr=args.learning_rate,
                        weight_decay=args.l2_reg)

  # num_batches = ceil(len(train_set.dataset) / float(bsz))
  outer_loss_list = []

  epoch = 0
  # while time() - start < args.n_minutes * 60:
  while epoch < args.n_epochs:
    model.train()

    epoch += 1
    print('\t' * 3, "Rank %d - Epoch %d" % (rank, epoch))

    # ----------------------------------------
    #                   SGD
    # ----------------------------------------
    epoch_loss = 0.0
    c = 0
    for data, target in train_set:
      c += 1
      data, target = Variable(data), Variable(target)

      optimizer.zero_grad()
      loss = calculate_model_loss(model, data, target)
      epoch_loss += loss.data
      loss.backward()

      # synchronization barrier
      average_gradients(model)

      # SGD update
      optimizer.step()

    # ----------------------------------------
    #                  REPORT
    # ----------------------------------------
    # normalize loss locally - good for skewed data
    epoch_loss = epoch_loss / c
    print("RANK={} LOSS={} N_BAT={} size={}".format(rank, epoch_loss, c,
                                                    size))

    # all to master node
    dist.reduce(epoch_loss, 0, op=dist.ReduceOp.SUM)

    if rank == 0:
      print('Rank:', rank, ', epoch ', epoch, ': ', epoch_loss / size)
      current = datetime.datetime.fromtimestamp(time()).strftime(
        '%Y-%m-%d %H:%M:%S')
      write_to_file = os.path.join(output_dir, 'train_{}_{}_{}.txt'.
                                   format(desc, type, size))
      with open(write_to_file, 'a') as f:
        f.write('{}\t{}\n'.format(current, float(epoch_loss / size)))

    outer_loss_list.append(epoch_loss / size)

    # ----------------------------------------
    #             EVALUATION
    # ----------------------------------------
    test_dataset(rank, size, model, test_set, bsz,
                 master_rank=0, type=type, desc=desc, output_dir=output_dir)

  print(outer_loss_list)
