#!/usr/bin/env python

"""DISTRIBUTED SVRG"""

import datetime
import os
from random import choice
from random import randint
from time import time

import torch
import torch.distributed as dist
import torch.optim as optim
from torch.autograd import Variable

# local imports
from config import args
from utils import average_gradients
from utils import calculate_model_loss
from utils import get_current_gradients
from utils import get_current_params
from utils import init_dataset
from utils import init_model
from utils import init_zero_grads
from utils import set_current_grads
from utils import set_current_params
from utils import test_dataset


def dsvrg(rank, size):
  """ Distributed Synchronous SVRG  """
  n_inner_loops = args.n_inner_loops  # inner loop frequency
  type = args.model_type
  desc = 'svrg'

  output_dir = '../output_{}_{}'.format(args.dataset, type)
  if not os.path.exists(output_dir):
    os.mkdir(output_dir)

  # -----------------------------------
  #              DATASET
  # -----------------------------------
  train_set, bsz = init_dataset(args.dataset,
                                args.batch_size,
                                is_train=True,
                                rank=rank)
  test_set, bsz = init_dataset(args.dataset,
                               args.batch_size,
                               is_train=False,
                               rank=rank)

  # -----------------------------------
  #              MODEL
  # -----------------------------------
  model = init_model(args.dataset)
  model2 = init_model(args.dataset)

  optimizer = optim.SGD(model.parameters(), lr=args.learning_rate)
  optimizer2 = optim.SGD(model2.parameters(), lr=args.learning_rate)

  outer_loss_list = []

  # --------------------------------------------
  #               OUTER LOOP
  # --------------------------------------------
  epoch = 0
  while epoch < args.n_epochs:
    epoch += 1
    epoch_loss = 0.0

    # this var is a list of layer-wise weights (different-shape tensors)
    current_outer_grads = None

    c = 0
    # for each node, accumulate grads manually for all batches
    # and set this accumulated grads into the "model"
    for data, target in train_set:
      c += 1
      data, target = Variable(data), Variable(target)

      optimizer.zero_grad()
      loss = calculate_model_loss(model, data, target)
      epoch_loss += loss.data

      loss.backward()

      # take a snapshot of the current node's local gradient
      outer_batch_grads = get_current_gradients(model)


      # -------------------------------------------------------
      # TODO: accumulate current_outer_grads for all batches
      # -------------------------------------------------------

      current_outer_grads = None

      # -------------------------------------------------------



    # normalize outer grads after we have done with all batches in train_set
    current_outer_grads = [g / c for g in current_outer_grads]

    # take snapshot before averaging (node-wise)
    set_current_grads(model, current_outer_grads)




    # --------------------------------------------
    # TODO: Sync all nodes here,
    # get the global gradients
    # and save it into `avg_grads`
    # hint: see SGD
    # --------------------------------------------

    avg_grads = None

    # --------------------------------------------



    # normalize loss locally - good for skewed data
    epoch_loss = epoch_loss / c

    # --------------------------------------------
    #   (OUTER LOOP REPORT) SYNC FOR THIS EPOCH
    # --------------------------------------------
    dist.reduce(epoch_loss, 0, op=dist.ReduceOp.SUM)
    if rank == 0:
      print('Rank:', rank, ': Outer loop: Rank', dist.get_rank(),
            ', epoch ', epoch, ': ', (epoch_loss / size))
      current = datetime.datetime.fromtimestamp(time()). \
        strftime('%Y-%m-%d %H:%M:%S')
      write_to_file = os.path.join(output_dir,
                                   'train_{}_{}_{}.txt'.format(desc, type,
                                                               size))
      with open(write_to_file, 'a') as f:
        f.write('{}\t{}\n'.format(current, float(epoch_loss / size)))

    # this statistic is what we care about alongside accuracy (if in cases of
    # classification models)
    outer_loss_list.append(epoch_loss / size)





    # --------------------------------------------
    #                INNER LOOPS
    # --------------------------------------------
    # inner_loop_params will save all grads for all inner loop's results
    # hint: read the SVRG algorithm in Figure 1, option II
    grad_len, inner_loop_params = init_zero_grads(model,
                                                  n_times=n_inner_loops)

    # this set of params is for the initialization at the beginning of each
    # inner loop only
    _, [inner_init_params] = init_zero_grads(model)

    for inner in range(n_inner_loops):
      # assume node 0 is the master, it will randomly sample the chosen node
      # this is done repetitively for each inner loop
      chosen = torch.tensor(0)

      if rank == 0:
        chosen_rank = randint(0, size - 1)
        print("\tInner %d: At master rank = %d: Chosen rank is %d" %
              (inner, rank, chosen_rank))
        chosen = torch.tensor(chosen_rank)

      # after this, all nodes will know who is chosen
      dist.barrier()
      dist.all_reduce(chosen, op=dist.ReduceOp.SUM)
      dist.barrier()

      # ----------------------------------
      #         INIT at CHOSEN NODE
      # ----------------------------------
      print("INIT PHASE AT CHOSEN RANK: {}".format(int(chosen)))
      if inner == 0:
        if rank == int(chosen):
          w_t_0 = get_current_params(model)
          set_current_params(model2, w_t_0)
        elif rank == 0:
          w_t_0 = get_current_params(model)
          inner_loop_params[inner] = w_t_0
      # else: server has to send w_{t-1} to chosen rank
      else:
        if int(chosen) == 0:
          if rank == 0:
            set_current_params(model2, inner_loop_params[inner])
        else:

          # -------------------------------------------------------
          # TODO: send the last inner's weights to the chosen node
          # -------------------------------------------------------

          pass

          # -------------------------------------------------------




          # now node `chosen` has the from-server weights to begin this loop
          if rank == int(chosen):
            # overwrite params into rank `chosen`
            set_current_params(model2, inner_init_params)


      # now every one knows which node is chosen
      # let it finish the job
      inner_epoch_loss = 0.
      updated_params = None  # only chosen rank has this updated after inner loop

      if rank == int(chosen):
        for data, target in train_set:
          data, target = Variable(data), Variable(target)

          optimizer2.zero_grad()
          inner_loss = calculate_model_loss(model2, data, target)
          inner_epoch_loss += inner_loss.data



          # -------------------------------------------------------
          # TODO: as follows
          #   1. Calculate gradients for this batch
          #   2. Re-calculate gradients for this batch w.r.t the algorithms
          #       (Hint: make use of `avg_grads as well)
          #   3. Update SGD-like way for this batch
          # -------------------------------------------------------

          pass

          # -------------------------------------------------------

        updated_params = get_current_params(model2)

      # prepare for the next round init (inner + 1)
      if inner < n_inner_loops - 1:
        if int(chosen) == 0:
          if rank == 0:
            inner_loop_params[inner + 1] = updated_params.copy()
        else:
          for i in range(grad_len):
            if rank == int(chosen):
              dist.send(updated_params[i], dst=0)
            elif rank == 0:
              dist.recv(inner_loop_params[inner + 1][i], src=int(chosen))

    # ---------------------END OF ALL INNNER LOOPS------------------------------
    dist.barrier() # needed or not?


    # --------------------------------------------------------------------------
    # TODO: Now all inner loops have finished,
    #       we have `inner_loop_params` list at the master node.
    # Now do as follows:
    #   1. Sample randonly the chosen weights \tilde{w_s} for the next epoch
    #   2. Sync this chosen weights to all nodes: `chosen_params_next_epoch`
    #     (Hint: into `model`, not `model2`)
    # --------------------------------------------------------------------------

    # init all zeros for broadcasting
    # hint: why do we need this to be all zeros for all nodes at first
    _, [chosen_params_next_epoch] = init_zero_grads(model)

    pass

    # --------------------------------------------------------------------------



    # ----------------------------------------
    #              EVALUATION
    # ----------------------------------------
    test_dataset(rank, size, model, test_set, bsz,
                 master_rank=0, type=type, desc=desc, output_dir=output_dir)

  print(outer_loss_list)
