#!/usr/bin/env python

import random
from random import Random

import numpy as np
import torch
import torch.distributed as dist
from torch.utils.data import TensorDataset
from torchvision import datasets, transforms

from config import args


class Partition(object):
  """ Define how to split the dataset, optionally make it skewed """
  def __init__(self, data, index, is_skewed=False):
    self.data = data
    # self.index = index

    # make data
    labels = []
    self.index = []
    n_classes = 10

    torch.random.manual_seed(index[0])
    sample_probs = torch.rand(n_classes)

    # delele some classes to make it skewed - optional
    np.random.seed(index[0])
    to_delete = [np.random.randint(0, n_classes) for _ in range(2)]

    # only keep 1 class per node - another optional setting for skewed data
    chosen = dist.get_rank() % n_classes
    # print("Chosen = {}".format(chosen))

    for idx in index:
      (x, y) = data[idx]
      if is_skewed:
        # if random.random() < sample_probs[y]:
        #   if y not in to_delete:
        if y == chosen:
          # take this index
          self.index.append(idx)
          labels.append(y)
      else:
        self.index.append(idx)
        labels.append(y)

    print("Dataset len = {}, histogram={}".format(len(self.index),
                                                  np.histogram(labels)))

  def __len__(self):
    return len(self.index)

  def __getitem__(self, index):
    data_idx = self.index[index]
    return self.data[data_idx]


class DataPartitioner(object):
  """ Partitions a dataset into different chunks. """

  def __init__(self, data, sizes=[0.1]*10, seed=1234):
    """Sizes is a list of data portion in percentage of each node"""
    self.data = data
    self.partitions = []
    rng = Random()
    rng.seed(seed)
    data_len = len(data)
    indexes = [x for x in range(0, data_len)]

    rng.shuffle(indexes)

    for frac in sizes:
      part_len = int(frac * data_len)
      self.partitions.append(indexes[0:part_len])
      indexes = indexes[part_len:]

  def use(self, partition, is_skewed=False):
    return Partition(self.data, self.partitions[partition], is_skewed)


def partition_data_real(name='abalone', batch_size=128,
                        is_train=True, rank=None):
  """Note: param batch_size is not used, instead n_batches """
  if name == 'abalone':
    tail = 'train' if is_train else 'test'
    datapath = '../data/abalone/abalone.{}'.format(tail)

    n_machines = dist.get_world_size()
    n_labels_this_rank = 28 // n_machines
    selected_labels = list(range(rank * n_labels_this_rank,
                                 (rank + 1) * n_labels_this_rank))

    X, y = [], []
    with open(datapath) as f:
      for line in f:
        line = line.strip().split(',')

        # convert a line to numbers
        if line[0] == 'M':
          line[0] = 1
        elif line[0] == 'F':
          line[0] = 2
        else:
          line[0] = 3

        data = [float(x) for x in line[:-1]]
        target = float(line[-1])
        target = target - 1 if target != 29 else (target - 2)  # index form 0

        if target in selected_labels:
          X.append(data)
          y.append(target)

      X = np.array(X)
      y = np.array(y)

      x_tensor = torch.from_numpy(X).float()
      y_tensor = torch.from_numpy(y).float()

      dataset = TensorDataset(x_tensor, y_tensor)
      bsz = len(y) // args.n_batches + 1

      is_shuffled = True if is_train else False
      loader = torch.utils.data.DataLoader(dataset,
                                           batch_size=bsz,
                                           shuffle=is_shuffled)
  else:
    raise NotImplementedError

  return loader, bsz


def partition_train_mnist(batch_size=1280):
  """ Partitioning MNIST """
  random_seed = 54334
  np.random.seed(random_seed + 1000)
  torch.manual_seed(random_seed + 2000)
  random.seed(random_seed + 3000)

  dataset = datasets.MNIST(
    '../data',
    train=True,
    download=True,
    transform=transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize((0.1307,), (0.3081,))
    ]))

  size = dist.get_world_size()
  print("size = %d" % size)

  partition_sizes = [1.0 / size for _ in range(size)]
  partition = DataPartitioner(dataset, partition_sizes)
  partition = partition.use(dist.get_rank(), args.is_skewed)

  bsz = batch_size // size  # batch size for 1 node

  # optionally can use this
  # refined_len = len(partition.data)
  # bsz = refined_len // args.n_batches + 1

  train_loader = torch.utils.data.DataLoader(partition,
                                             batch_size=bsz,
                                             shuffle=True)
  return train_loader, bsz


def partition_train_cifar(batch_size=1280):
  """ Partitioning CIFAR10 """
  dataset = datasets.CIFAR10(
    '../data',
    train=True,
    download=True,
    transform=transforms.Compose([
      transforms.RandomCrop(32, padding=4),
      transforms.RandomHorizontalFlip(),
      transforms.ToTensor(),
      transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
    ])
  )

  size = dist.get_world_size()
  print("size = %d" % size)

  partition_sizes = [1.0 / size for _ in range(size)]
  partition = DataPartitioner(dataset, partition_sizes)
  partition = partition.use(dist.get_rank(), args.is_skewed)

  bsz = batch_size // size  # batch size for 1 node

  # optionally can use this
  # refined_len = len(partition.data)
  # bsz = refined_len // args.n_batchess + 1

  train_set = torch.utils.data.DataLoader(
    partition, batch_size=bsz, shuffle=True)

  return train_set, bsz


def partition_test_mnist(batch_size=1280):
  """ Partitioning MNIST """
  dataset = datasets.MNIST(
    '../data',
    train=False,
    download=True,
    transform=transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize((0.1307,), (0.3081,))
    ]))

  size = dist.get_world_size()
  print("size = %d" % size)

  partition_sizes = [1.0 / size for _ in range(size)]
  partition = DataPartitioner(dataset, partition_sizes)
  partition = partition.use(dist.get_rank())

  bsz = batch_size // size  # batch size for 1 node

  # optionally can use this
  # refined_len = len(partition.data)
  # bsz = refined_len // args.n_batchess + 1

  test_set = torch.utils.data.DataLoader(
    partition, batch_size=bsz, shuffle=False)

  return test_set, bsz


def partition_test_cifar(batch_size=1280):
  """ Partitioning CIFAR10 """
  dataset = datasets.CIFAR10(
    '../data',
    train=False,
    download=True,
    transform=transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
    ])
  )

  size = dist.get_world_size()
  print("size = %d" % size)

  partition_sizes = [1.0 / size for _ in range(size)]
  partition = DataPartitioner(dataset, partition_sizes)
  partition = partition.use(dist.get_rank())

  bsz = batch_size // size  # batch size for 1 node

  # optionally can use this
  # refined_len = len(partition.data)
  # bsz = refined_len // args.n_batches + 1

  test_set = torch.utils.data.DataLoader(
    partition, batch_size=bsz, shuffle=True)

  return test_set, bsz
