#!/usr/bin/env python

import datetime
import os
from time import time

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from config import args
from torch.autograd import Variable

from data import partition_data_real
from data import partition_test_cifar
from data import partition_test_mnist
from data import partition_train_cifar
from data import partition_train_mnist
from models import CIFAR10_CNN
from models import CIFAR10_CONVEX
from models import MNIST_CNN
from models import MNIST_CONVEX
from models import Regression_Real


def set_current_grads(model, new_grads, divide=1.):
  """
  Set Gradients for a model, with an option to average by a divide factor
  
  Args:
    model: target model graph 
    new_grads: to overwite to this model  
    divide: average factor 

  Returns: nothing, it's in-place operation, the model data is 
            changed after this 

  """
  assert divide > 0
  for idx, param in enumerate(model.parameters()):
    param.grad = new_grads[idx] / divide


def get_current_gradients(model):
  """ Get current grads of a model"""

  current_params_grads = []
  for param in model.parameters():
    current_params_grads.append(param.grad)

  return current_params_grads


def set_current_params(model, new_params, divide=1.):
  """ Same as for gradients above, but for params """
  assert divide > 0
  for idx, param in enumerate(model.parameters()):
    param.data = new_params[idx] / divide


def get_current_params(model):
  """ Same as for gradients above, but for params """

  current_params_values = []
  for param in model.parameters():
    current_params_values.append(param.data)

  return current_params_values


def init_zero_grads(model, n_times=1):
  """
  Init gradients all zeros for a model
     Can also be used for PARAMS since they have the exactly the same shapes 
     
  Args:
    model: an exemplar one to extract scaffold of params/grads 
    n_times: how many identical elements in the returned list 

  Returns: a list of model gradients or params (of how many times we want)

  """
  zero_grads = []
  for param in model.parameters():
    zero_grads.append(torch.zeros_like(param.grad))

  return len(zero_grads), [zero_grads for _ in range(n_times)]


def average_gradients(model):
  """ Gradient averaging for all nodes (using `all_reduce` operation) """
  current_params_values = []
  current_params_grads = []

  for param in model.parameters():
    dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM)

    param.grad.data /= dist.get_world_size()  # global var (also dist.get_world_size())

    current_params_values.append(param.data)
    current_params_grads.append(param.grad)

  return current_params_values, current_params_grads


def average_params(model):
  """ Same as for gradients above, but for params"""
  current_params_values = []
  current_params_grads = []

  size = float(dist.get_world_size())
  for param in model.parameters():
    dist.all_reduce(param.data.data, op=dist.ReduceOp.SUM)
    param.data.data /= size

    current_params_values.append(param.data)
    current_params_grads.append(param.grad)

  return current_params_values, current_params_values


def average_grads_and_params(model):
  """ Do averaging for both grads and params at the same time """
  current_params_values = []
  current_params_grads = []

  size = float(dist.get_world_size())
  for param in model.parameters():
    # dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM, group=0)
    dist.all_reduce(param.data.data, op=dist.ReduceOp.SUM)
    param.data.data /= size

    dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM)
    param.grad.data /= size

    current_params_values.append(param.data)
    current_params_grads.append(param.grad)

  return current_params_values, current_params_values


def calculate_model_loss(model, data, target):
  output = model(data)
  criterion = F.mse_loss if args.model_type == 'regression' else F.nll_loss
  loss = criterion(output, target)

  return loss


def calculate_loss_value(rank, size, model, data_set, batch_size,
                         master_rank=0, type='convex', desc=None,
                         output_dir=None):
  """
  Calculate loss value for each node 
  
  Args:
    rank: 
    size: 
    model: associated with each model, of which graph is the same  
    data_set: 
    batch_size: 
    master_rank: default is 0 
    type: model_type in args
    desc: see the main function of each algorithm 
    output_dir: see the main function of each algorithm 

  """
  model.eval()

  loss = 0.
  # num_batches = ceil(len(data_set.dataset) / float(batch_size))

  c = 0
  for data, target in data_set:
    c += 1
    data, target = Variable(data), Variable(target)

    loss += calculate_model_loss(model, data, target).data

  # noralize loss by n_batches
  loss = loss / c
  print("RANK={} LOSS={} n-batches={} size={}".format(rank, loss, c, size))

  # dist.barrier()
  dist.reduce(loss, master_rank, op=dist.ReduceOp.SUM)

  if rank == master_rank:
    # print("o" * 80)
    # print('Rank:', rank,
    #       ': INIT LOSS : Rank', dist.get_rank(),
    #       ': ', loss / size)

    current = datetime.datetime.fromtimestamp(time()).strftime(
      '%Y-%m-%d %H:%M:%S')
    write_to_file = os.path.join(output_dir,
                                 'train_{}_{}_{}.txt'.format(desc, type, size))
    with open(write_to_file, 'a') as f:
      f.write('{}\t{}\n'.format(current, float(loss / size)))
      # print("o" * 100)


def test_dataset(rank, size, model, test_set, batch_size,
                 master_rank=0, type='convex', desc=None, output_dir=None):
  """
  Test dataset, no need to change   
  """
  print("Training is done, testing now with len {}....".
        format(len(test_set.dataset)))

  model.eval()

  test_loss = 0.
  correct = 0.
  # num_batches = ceil(len(test_set.dataset) / float(batch_size))
  # print("Num batches = {}".format(num_batches))

  c = 0
  for data, target in test_set:
    c += 1
    data, target = Variable(data), Variable(target)

    if args.model_type == 'regression':
      loss = calculate_model_loss(model, data, target)
      test_loss += loss.data
    else:
      output = model(data)
      loss = F.nll_loss(output, target)

      test_loss += loss.data
      pred = output.data.max(1, keepdim=True)[1]
      correct += pred.eq(target.data.view_as(pred)).cpu().sum()

  # normalize loss locally - good for skewed data
  test_loss = test_loss / c

  # sync
  dist.barrier()
  dist.reduce(test_loss, master_rank, op=dist.ReduceOp.SUM)
  if args.model_type != 'regression':
    dist.reduce(correct, master_rank, op=dist.ReduceOp.SUM)

  print("Rank {} correct = {}".format(rank, correct))

  if rank == master_rank:
    if not os.path.exists(output_dir):
      os.mkdir(output_dir)
    current = datetime.datetime.fromtimestamp(time()). \
      strftime('%Y-%m-%d %H:%M:%S')
    write_to_file = os.path.join(output_dir,
                                 'test_{}_{}_{}.txt'.format(desc, type, size))
    with open(write_to_file, 'a') as f:
      f.write('{}\t{}\t{}\n'.
              format(current,
                     float(test_loss / size),
                     float(100. * correct / size / len(test_set.dataset))))
    print('\nTest Average loss: {:.4f}, Accuracy: {}/{} ({:.1f}%)\n'
          .format(test_loss / size,
                  correct,
                  size * len(test_set.dataset),
                  100. * correct / size / len(test_set.dataset)))


def init_linear_weights(m):
  """init fc layers with glorot uniform"""
  if type(m) == nn.Linear:
    nn.init.xavier_uniform_(m.weight)
    m.bias.data.fill_(0.01)


def init_conv2d_weights(m):
  """init cnn layers with glorot uniform"""
  if isinstance(m, nn.Conv2d):
    nn.init.xavier_uniform_(m.weight)
    nn.init.zeros_(m.bias)


def init_dataset(name, batch_size, is_train=True, rank=None):
  """
    Split data and distribute to nodes in the cluster 
  """
  if name == 'mnist':
    if is_train:
      return partition_train_mnist(batch_size)
    else:
      return partition_test_mnist(batch_size)
  elif name == 'cifar10':
    if is_train:
      return partition_train_cifar(batch_size)
    else:
      return partition_test_cifar(batch_size)
  elif name == 'abalone':
    return partition_data_real(name, batch_size, is_train, rank)
  else:
    raise NotImplementedError


def init_model(dataset):
  """Init a single model and load init weights uniformly each time"""
  if dataset == 'mnist':
    if args.model_type == 'convex':
      model = MNIST_CONVEX()
      model.load_state_dict(torch.load(args.mnist_convex_weights))
    elif args.model_type == 'cnn':
      model = MNIST_CNN()
      model.load_state_dict(torch.load(args.mnist_cnn_weights))
    else:
      raise NotImplementedError
  elif dataset == 'cifar10':
    if args.model_type == 'convex':
      model = CIFAR10_CONVEX()
      model.load_state_dict(torch.load(args.cifar10_convex_weights))
    elif args.model_type == 'cnn':
      model = CIFAR10_CNN()
      model.load_state_dict(torch.load(args.cifar10_cnn_weights))
    else:
      raise NotImplementedError
  elif dataset == 'abalone':
    if args.model_type == 'regression':
      model = Regression_Real(8)
      model.load_state_dict(torch.load(args.abalone_regression_weights))
    else:
      raise NotImplementedError
  else:
    raise NotImplementedError

  return model


def init_processes(rank, size, fn, backend='gloo', master_addr=None, port=None):
  """ Initialize the distributed environment. 
      For massive tests on the same node, 
        change this MASTER PORT for each test.
  """
  os.environ['MASTER_ADDR'] = \
    '127.0.0.1' if master_addr is None else str(master_addr)
  os.environ['MASTER_PORT'] = '29500' if port is None else str(port)
  dist.init_process_group(backend, rank=rank, world_size=size)
  fn(rank, size)
